/*
  $Id: statusbarHud.c,v 1.48 2012/03/31 19:15:16 dug9 Exp $

*/

/****************************************************************************
    This file is part of the FreeWRL/FreeX3D Distribution.

    Copyright 2009 CRC Canada. (http://www.crc.gc.ca)

    FreeWRL/FreeX3D is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FreeWRL/FreeX3D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FreeWRL/FreeX3D.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************************/


#include <config.h>
#include <system.h>
#include <display.h>
#include <internal.h>

#include <libFreeWRL.h>
#include "main/headers.h"
#include "vrml_parser/Structs.h"
#include "scenegraph/Viewer.h"
#include "opengl/Textures.h"
#include "opengl/LoadTextures.h"
#include "main/MainLoop.h"

/* the following are bitmap icons for the toolbar, 
generated by writing out C structs from thresholded png icons. Setting 2
parameters in this statusbarHud.c causes it to read in your 32x32xRGBA .pngs.
and write out C struct versions: 
buttonType = 0; // 0 = rgba .png 1= .c bitmap (see above) 
savePng2dotc = 1; // if you read png and want to save to a bitmap .c struct, put 1 
*/ 
#if !(defined(IPHONE) || defined(_ANDROID)) // || defined(GLES2))

GLubyte walk[] = {
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x3,0x80,0x0,0x0,
0x7,0xc0,0x0,0x0,
0x7,0xf0,0x0,0x0,
0x7,0xf8,0x0,0x0,
0x3,0xfe,0x0,0x0,
0x1,0xff,0x80,0x0,
0x0,0xff,0xc0,0x0,
0x0,0x7f,0xf0,0x0,
0x0,0x7f,0xf8,0x0,
0x0,0x3f,0xfe,0x0,
0x0,0x1f,0xff,0x80,
0x0,0xf,0xff,0x80,
0x0,0x7,0xff,0x0,
0x0,0x7,0xfe,0x0,
0x0,0x3,0xfc,0x0,
0x0,0x1,0xf8,0xc0,
0x0,0x0,0xf0,0x0,
0x0,0x0,0x63,0x0,
0x0,0x0,0x43,0x0,
0x0,0x0,0x1c,0x0,
0x0,0x0,0x1e,0x0,
0x0,0x0,0xe,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
};
GLubyte fly[] = {
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x80,
0x0,0x0,0x0,0xc0,
0x0,0x80,0x0,0xc0,
0x1,0xc0,0x0,0xc0,
0x3,0xc0,0x1,0xc0,
0x3,0xe0,0x3,0xc0,
0x7,0xf0,0x1f,0xc0,
0xf,0xfd,0xff,0xc0,
0x1,0xff,0xff,0xe0,
0x0,0x7f,0xff,0xc0,
0x0,0x3f,0xff,0x0,
0x0,0x1f,0xf8,0x0,
0x0,0x1f,0xf8,0x0,
0x0,0x1f,0xfc,0x0,
0x0,0x1f,0xfc,0x0,
0x0,0x1f,0xfe,0x0,
0x0,0x1f,0xff,0x0,
0x0,0x1f,0xff,0x80,
0x0,0x1f,0xf,0xe0,
0x0,0x1f,0x3,0x80,
0x0,0x1f,0x0,0x0,
0x0,0x3f,0x0,0x0,
0x1,0xff,0x0,0x0,
0x3,0xff,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
};
GLubyte flyex[] = {
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x80,
0x0,0x0,0x40,0xc0,
0x0,0x80,0x40,0xc0,
0x1,0xc6,0x40,0xc0,
0x3,0xc3,0xc1,0xc0,
0x3,0xe1,0x83,0xc0,
0x7,0xf1,0x1f,0xc0,
0xf,0xff,0xff,0xc0,
0x1,0xff,0xff,0xe0,
0x0,0x7f,0xff,0xc0,
0x0,0x3f,0xff,0x0,
0x0,0x1f,0xf8,0x0,
0x0,0x1f,0xf8,0x0,
0x0,0x1f,0xfc,0x0,
0x0,0x1f,0xfc,0x0,
0x0,0x1f,0xfe,0x0,
0x0,0x1f,0xff,0x0,
0x0,0x1f,0xff,0x80,
0x0,0x1f,0xf,0xe0,
0x0,0x1f,0x3,0x80,
0x0,0x1f,0x0,0x0,
0x0,0x3f,0x0,0x0,
0x1,0xff,0x0,0x0,
0x3,0xff,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
};
GLubyte examine[] = {
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x7c,0x0,0x0,
0x0,0xfe,0x0,0x0,
0x1,0xff,0x0,0x0,
0x1,0x83,0x0,0x0,
0x3,0x83,0x87,0x0,
0x3,0x83,0x87,0xe0,
0x1,0x87,0x0,0xe0,
0x1,0xff,0x0,0x0,
0x0,0xfe,0x1c,0x0,
0x0,0x7c,0x1f,0x0,
0x0,0x0,0x7,0x80,
0x0,0x0,0x61,0xe0,
0x1,0x0,0x70,0xf0,
0x3,0x80,0x78,0x30,
0x3,0x3,0x1c,0x0,
0x3,0x3,0x8e,0x0,
0x3,0x3,0x87,0x0,
0x0,0x1,0xc3,0x80,
0x0,0x0,0xe1,0xc0,
0x0,0x0,0xe0,0xe0,
0x0,0x0,0x70,0x40,
0x0,0x0,0x38,0x0,
0x0,0x0,0x38,0x0,
0x0,0x0,0x18,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
};
GLubyte headlight[] = {
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x1,0x80,0x0,0x0,
0x3,0xc0,0x0,0x0,
0x7,0xe0,0x0,0x0,
0x7,0xf0,0x0,0x0,
0x3,0xf8,0x0,0x0,
0x1,0xfc,0x0,0x0,
0x0,0xfe,0x0,0x0,
0x0,0x7f,0x0,0x0,
0x0,0x3f,0x80,0x0,
0x0,0x1f,0xc0,0x0,
0x0,0xf,0xe0,0x0,
0x0,0x7,0xf0,0x0,
0x0,0x3,0xf8,0x0,
0x0,0x1,0xfc,0x0,
0x0,0x0,0xff,0xc0,
0x0,0x0,0x7f,0xe0,
0x0,0x0,0x3f,0xe0,
0x0,0x0,0x1f,0xc0,
0x0,0x0,0x1f,0x80,
0x0,0x0,0x1f,0x0,
0x0,0x0,0x1e,0x0,
0x0,0x0,0xc,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
};
GLubyte collision[] = {
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x3,0xdf,0xfb,0xe0,
0x3,0xdf,0xfb,0xe0,
0x3,0xdf,0xfb,0xe0,
0x3,0x8f,0xf1,0xc0,
0x0,0x0,0x0,0x0,
0x3,0xfe,0x7f,0xc0,
0x7,0xfe,0x7f,0xe0,
0x7,0xfe,0x7f,0xe0,
0x3,0xfe,0x7f,0xe0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x3,0xdf,0xfb,0xe0,
0x3,0xdf,0xfb,0xe0,
0x3,0xdf,0xfb,0xe0,
0x3,0x8f,0xf1,0xc0,
0x0,0x0,0x0,0x0,
0x3,0xfe,0x7f,0xc0,
0x7,0xfe,0x7f,0xe0,
0x7,0xfe,0x7f,0xe0,
0x3,0xfe,0x7f,0xc0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
};
GLubyte prev[] = {
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x38,
0x0,0x0,0x0,0xf8,
0x0,0x0,0x3,0xf8,
0x0,0x0,0xf,0xf8,
0x0,0x0,0x3f,0xf8,
0x0,0x0,0xff,0xf8,
0x0,0x3,0xff,0xf8,
0x0,0xf,0xff,0xf8,
0x0,0x3f,0xff,0xf8,
0x0,0xff,0xff,0xf8,
0x3,0xff,0xff,0xf8,
0xf,0xff,0xff,0xf8,
0xf,0xff,0xff,0xf8,
0x3,0xff,0xff,0xf8,
0x0,0xff,0xff,0xf8,
0x0,0x3f,0xff,0xf8,
0x0,0xf,0xff,0xf8,
0x0,0x3,0xff,0xf8,
0x0,0x0,0xff,0xf8,
0x0,0x0,0x3f,0xf8,
0x0,0x0,0xf,0xf8,
0x0,0x0,0x3,0xf8,
0x0,0x0,0x0,0xf8,
0x0,0x0,0x0,0x38,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
};
GLubyte next[] = {
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x1c,0x0,0x0,0x0,
0x1f,0x0,0x0,0x0,
0x1f,0xc0,0x0,0x0,
0x1f,0xf0,0x0,0x0,
0x1f,0xfc,0x0,0x0,
0x1f,0xff,0x0,0x0,
0x1f,0xff,0xc0,0x0,
0x1f,0xff,0xf0,0x0,
0x1f,0xff,0xfc,0x0,
0x1f,0xff,0xff,0x0,
0x1f,0xff,0xff,0xc0,
0x1f,0xff,0xff,0xf0,
0x1f,0xff,0xff,0xf0,
0x1f,0xff,0xff,0xc0,
0x1f,0xff,0xff,0x0,
0x1f,0xff,0xfc,0x0,
0x1f,0xff,0xf0,0x0,
0x1f,0xff,0xc0,0x0,
0x1f,0xff,0x0,0x0,
0x1f,0xfc,0x0,0x0,
0x1f,0xf0,0x0,0x0,
0x1f,0xc0,0x0,0x0,
0x1f,0x0,0x0,0x0,
0x1c,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
};
GLubyte help[] = {
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0xf,0xe0,0x0,
0x0,0x1f,0xf0,0x0,
0x0,0x1f,0xf0,0x0,
0x0,0x7,0xc0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x4,0x0,0x0,
0x0,0x3e,0x0,0x0,
0x0,0x3e,0x0,0x0,
0x0,0x3f,0x80,0x0,
0x0,0x3f,0xc0,0x0,
0x0,0x1f,0xe0,0x0,
0x0,0xf,0xf8,0x0,
0x0,0x7,0xfc,0x0,
0x0,0x1,0xff,0x0,
0x0,0x0,0xff,0x80,
0x0,0x0,0x3f,0xc0,
0x0,0x0,0x3f,0xc0,
0x0,0x0,0x3f,0xc0,
0x0,0x0,0x3f,0xc0,
0x0,0x13,0xff,0xc0,
0x0,0x1f,0xff,0x0,
0x0,0xf,0xfc,0x0,
0x0,0x3,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
};
GLubyte messages[] = {
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x3,0xc0,0x0,
0x0,0xf,0xf0,0x0,
0x0,0xf,0xf0,0x0,
0x0,0x7,0xe0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x1,0x80,0x0,
0x0,0x3,0xc0,0x0,
0x0,0x3,0xc0,0x0,
0x0,0x3,0xc0,0x0,
0x0,0x7,0xc0,0x0,
0x0,0x7,0xe0,0x0,
0x0,0x7,0xe0,0x0,
0x0,0x7,0xe0,0x0,
0x0,0xf,0xe0,0x0,
0x0,0xf,0xf0,0x0,
0x0,0xf,0xf0,0x0,
0x0,0xf,0xf0,0x0,
0x0,0xf,0xf0,0x0,
0x0,0x7,0xe0,0x0,
0x0,0x7,0xe0,0x0,
0x0,0x3,0xc0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
};
GLubyte options[] = {
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x1f,0xff,0xe0,0x0,
0x1f,0xff,0xe0,0x0,
0x18,0x20,0x60,0x0,
0x18,0x30,0x60,0x0,
0x18,0x78,0x60,0x0,
0x18,0x78,0x60,0x0,
0x18,0x7c,0x60,0x0,
0x18,0xfe,0x60,0x0,
0x18,0xef,0xe0,0x0,
0x19,0xc7,0xe0,0x0,
0x1b,0xc3,0xe0,0x0,
0x19,0x81,0xe0,0x0,
0x18,0x0,0xf0,0x0,
0x18,0x0,0x7c,0x0,
0x1f,0xff,0xfe,0x0,
0x1f,0xff,0xff,0x0,
0x0,0x0,0x7,0xc0,
0x0,0x0,0x3,0xe0,
0x0,0x0,0x1,0xf0,
0x0,0x0,0x0,0xf8,
0x0,0x0,0x0,0x38,
0x0,0x0,0x0,0x10,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
};
GLubyte level[] = {
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0xf,0xff,0xff,0xf0,
0x1f,0xff,0xff,0xf8,
0x1f,0xff,0xff,0xf8,
0x1f,0xff,0xff,0xf8,
0x1f,0xfc,0x1f,0xf8,
0x1f,0xf0,0xf,0xf8,
0x1f,0xe0,0x7,0xf8,
0x1f,0xe0,0x7,0xf0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
};
GLubyte reload[] = {
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x38,0x0,0x0,
0x0,0x78,0x0,0x0,
0x0,0xf0,0xf,0x80,
0x3,0xf0,0x3f,0xc0,
0x3,0xf0,0x3f,0xc0,
0x7,0xf8,0x3f,0xe0,
0x7,0xfc,0x3f,0xf0,
0xf,0xfc,0x3f,0xf0,
0xf,0xf8,0x1f,0xf0,
0x1f,0xf0,0xf,0xf8,
0x1f,0xe0,0xf,0xf8,
0x1f,0xe0,0x7,0xf8,
0x1f,0xe0,0x7,0xf8,
0x1f,0xe0,0xf,0xf8,
0x1f,0xf0,0xf,0xf8,
0xf,0xf8,0x1f,0xf0,
0xf,0xfe,0x7f,0xf0,
0x7,0xff,0xff,0xf0,
0x7,0xff,0xff,0xe0,
0x3,0xff,0xff,0xc0,
0x1,0xff,0xff,0xc0,
0x0,0xff,0xff,0x0,
0x0,0x7f,0xfe,0x0,
0x0,0x1f,0xfc,0x0,
0x0,0x7,0xc0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
};
/* <<< bitmap menu button icons */

/*fw fixed size bitmap fonts >>>
first 1: char # 0-127 (ie '!'=33, 'a'=97)
next 6: FW_GL_BITMAP(width,height,xbo,ybo,xadv,yadv,
last 7+: FW_GL_BITMAP(,,,,,,const GLubyte *bitmap);
*/
GLubyte fwLetters7x12[][19] = {
{32,7,12,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},
{33,7,12,0,0,7,0,0x0,0x0,0x40,0x40,0x0,0x40,0x40,0x40,0x40,0x40,0x0,0x0},
{34,7,12,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x50,0x50,0x50,0x50,0x0},
{35,7,12,0,0,7,0,0x0,0x0,0x50,0x28,0xfc,0x28,0x28,0x7e,0x24,0x12,0x0,0x0},
{36,7,12,0,0,7,0,0x0,0x20,0x70,0xa8,0x28,0x70,0xa0,0xa8,0x70,0x20,0x0,0x0},
{37,7,12,0,0,7,0,0x0,0x0,0x9c,0x54,0x54,0x2c,0xd0,0xa8,0xa8,0xe6,0x0,0x0},
{38,7,12,0,0,7,0,0x0,0x0,0x64,0x98,0x8c,0x50,0x20,0x58,0x48,0x30,0x0,0x0},
{39,7,12,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x20,0x10,0x10,0x0},
{40,7,12,0,0,7,0,0x10,0x20,0x20,0x40,0x40,0x40,0x40,0x20,0x20,0x10,0x0,0x0},
{41,7,12,0,0,7,0,0x0,0x40,0x20,0x20,0x10,0x10,0x10,0x20,0x20,0x40,0x0,0x0},
{42,7,12,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x20,0xa8,0x50,0xa8,0x20,0x0,0x0},
{43,7,12,0,0,7,0,0x0,0x0,0x0,0x8,0x8,0x3e,0x8,0x8,0x0,0x0,0x0,0x0},
{44,7,12,0,0,7,0,0x0,0x8,0x8,0x18,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},
{45,7,12,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0xf8,0x0,0x0,0x0,0x0,0x0,0x0},
{46,7,12,0,0,7,0,0x0,0x0,0x18,0x18,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},
{47,7,12,0,0,7,0,0x0,0x80,0x80,0x40,0x40,0x20,0x20,0x10,0x10,0x8,0x8,0x0},
{48,7,12,0,0,7,0,0x0,0x0,0x70,0x88,0x88,0xc8,0x98,0x88,0x88,0x70,0x0,0x0},
{49,7,12,0,0,7,0,0x0,0x0,0x38,0x10,0x10,0x10,0x10,0x10,0x30,0x10,0x0,0x0},
{50,7,12,0,0,7,0,0x0,0x0,0xf8,0x80,0x40,0x20,0x10,0x88,0x88,0x70,0x0,0x0},
{51,7,12,0,0,7,0,0x0,0x0,0x70,0x88,0x8,0x8,0x30,0x8,0x8,0xf8,0x0,0x0},
{52,7,12,0,0,7,0,0x0,0x0,0x10,0x10,0x10,0xf8,0x90,0x50,0x50,0x50,0x0,0x0},
{53,7,12,0,0,7,0,0x0,0x0,0xf0,0x8,0x8,0x88,0xf0,0x80,0x80,0xf8,0x0,0x0},
{54,7,12,0,0,7,0,0x0,0x0,0x70,0x88,0x88,0xc8,0xb0,0x80,0x80,0x78,0x0,0x0},
{55,7,12,0,0,7,0,0x0,0x0,0x20,0x20,0x20,0x20,0x10,0x10,0x88,0xf8,0x0,0x0},
{56,7,12,0,0,7,0,0x0,0x0,0x70,0x88,0x88,0x88,0x70,0x88,0x88,0x70,0x0,0x0},
{57,7,12,0,0,7,0,0x0,0x0,0x70,0x88,0x8,0x68,0x98,0x88,0x88,0x70,0x0,0x0},
{58,7,12,0,0,7,0,0x0,0x0,0x80,0x80,0x0,0x0,0x80,0x80,0x0,0x0,0x0,0x0},
{59,7,12,0,0,7,0,0x0,0x20,0x10,0x30,0x0,0x0,0x10,0x10,0x0,0x0,0x0,0x0},
{60,7,12,0,0,7,0,0x0,0x8,0x10,0x20,0x40,0x80,0x40,0x20,0x10,0x8,0x0,0x0},
{61,7,12,0,0,7,0,0x0,0x0,0x0,0x0,0xf8,0x0,0xf8,0x0,0x0,0x0,0x0,0x0},
{62,7,12,0,0,7,0,0x0,0x80,0x40,0x20,0x10,0x8,0x10,0x20,0x40,0x80,0x0,0x0},
{63,7,12,0,0,7,0,0x0,0x20,0x20,0x0,0x20,0x10,0x10,0x8,0x88,0x70,0x0,0x0},
{64,7,12,0,0,7,0,0x0,0x0,0x38,0x40,0x98,0xa4,0xa4,0x98,0x48,0x30,0x0,0x0},
{65,7,12,0,0,7,0,0x0,0x0,0x88,0xf8,0x88,0x88,0x50,0x50,0x20,0x20,0x0,0x0},
{66,7,12,0,0,7,0,0x0,0x0,0xf0,0x88,0x88,0x90,0xf0,0x88,0x88,0xf0,0x0,0x0},
{67,7,12,0,0,7,0,0x0,0x0,0x70,0x88,0x88,0x80,0x80,0x80,0x88,0x70,0x0,0x0},
{68,7,12,0,0,7,0,0x0,0x0,0xf0,0x88,0x88,0x88,0x88,0x88,0x88,0xf0,0x0,0x0},
{69,7,12,0,0,7,0,0x0,0x0,0xf8,0x80,0x80,0x80,0xe0,0x80,0x80,0xf8,0x0,0x0},
{70,7,12,0,0,7,0,0x0,0x0,0x80,0x80,0x80,0x80,0xe0,0x80,0x80,0xf8,0x0,0x0},
{71,7,12,0,0,7,0,0x0,0x0,0x68,0x98,0x88,0x88,0x90,0x80,0x88,0x78,0x0,0x0},
{72,7,12,0,0,7,0,0x0,0x0,0x88,0x88,0x88,0x88,0xf8,0x88,0x88,0x88,0x0,0x0},
{73,7,12,0,0,7,0,0x0,0x0,0x70,0x20,0x20,0x20,0x20,0x20,0x20,0x70,0x0,0x0},
{74,7,12,0,0,7,0,0x0,0x0,0x70,0x88,0x8,0x8,0x8,0x8,0x8,0x1c,0x0,0x0},
{75,7,12,0,0,7,0,0x0,0x0,0x84,0x88,0x90,0xa0,0xc0,0xa0,0x90,0x88,0x0,0x0},
{76,7,12,0,0,7,0,0x0,0x0,0xfc,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x0,0x0},
{77,7,12,0,0,7,0,0x0,0x0,0x88,0xa8,0xa8,0xa8,0xd8,0xd8,0x88,0x88,0x0,0x0},
{78,7,12,0,0,7,0,0x0,0x0,0x88,0x98,0x98,0xa8,0xa8,0xc8,0xc8,0x88,0x0,0x0},
{79,7,12,0,0,7,0,0x0,0x0,0x70,0x88,0x88,0x88,0x88,0x88,0x88,0x70,0x0,0x0},
{80,7,12,0,0,7,0,0x0,0x0,0x80,0x80,0x80,0xc0,0xb0,0x88,0x88,0xf0,0x0,0x0},
{81,7,12,0,0,7,0,0xc,0x10,0x70,0xa8,0xa8,0x88,0x88,0x88,0x88,0x70,0x0,0x0},
{82,7,12,0,0,7,0,0x0,0x0,0x84,0x88,0x90,0xa0,0xf0,0x88,0x88,0xf0,0x0,0x0},
{83,7,12,0,0,7,0,0x0,0x0,0x70,0x88,0x8,0x10,0x70,0x80,0x88,0x70,0x0,0x0},
{84,7,12,0,0,7,0,0x0,0x0,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0xf8,0x0,0x0},
{85,7,12,0,0,7,0,0x0,0x0,0x70,0x88,0x88,0x88,0x88,0x88,0x88,0x88,0x0,0x0},
{86,7,12,0,0,7,0,0x0,0x0,0x20,0x20,0x50,0x50,0x50,0x88,0x88,0x88,0x0,0x0},
{87,7,12,0,0,7,0,0x0,0x0,0x50,0x50,0xf8,0xa8,0xa8,0xa8,0x88,0x88,0x0,0x0},
{88,7,12,0,0,7,0,0x0,0x0,0x88,0x88,0x50,0x20,0x20,0x50,0x88,0x88,0x0,0x0},
{89,7,12,0,0,7,0,0x0,0x0,0x20,0x20,0x20,0x50,0x50,0x50,0x88,0x88,0x0,0x0},
{90,7,12,0,0,7,0,0x0,0x0,0xf8,0x80,0x40,0x20,0x20,0x10,0x8,0xf8,0x0,0x0},
{91,7,12,0,0,7,0,0x0,0x70,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x70,0x0,0x0},
{92,7,12,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},
{93,7,12,0,0,7,0,0x0,0x70,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x70,0x0,0x0},
{94,7,12,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x88,0x50,0x20,0x0,0x0},
{95,7,12,0,0,7,0,0x0,0xfe,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},
{96,7,12,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20,0x40,0x80,0x80,0x0},
{97,7,12,0,0,7,0,0x0,0x0,0x68,0x98,0x98,0x68,0x8,0x70,0x0,0x0,0x0,0x0},
{98,7,12,0,0,7,0,0x0,0x0,0x30,0xc8,0x88,0x88,0xc8,0xb0,0x80,0x80,0x0,0x0},
{99,7,12,0,0,7,0,0x0,0x0,0x70,0x88,0x80,0x80,0x88,0x70,0x0,0x0,0x0,0x0},
{100,7,12,0,0,7,0,0x0,0x0,0x68,0x98,0x88,0x98,0x98,0x68,0x8,0x8,0x0,0x0},
{101,7,12,0,0,7,0,0x0,0x0,0x70,0xc8,0x80,0xf8,0x88,0x70,0x0,0x0,0x0,0x0},
{102,7,12,0,0,7,0,0x0,0x0,0x40,0x40,0x40,0x40,0xf0,0x40,0x48,0x30,0x0,0x0},
{103,7,12,0,0,7,0,0x70,0x8,0x68,0x98,0x88,0x98,0x98,0x60,0x0,0x0,0x0,0x0},
{104,7,12,0,0,7,0,0x0,0x0,0x88,0x88,0x88,0x88,0xd0,0xb0,0x80,0x80,0x0,0x0},
{105,7,12,0,0,7,0,0x0,0x0,0x20,0x20,0x20,0x20,0x20,0x60,0x0,0x20,0x0,0x0},
{106,7,12,0,0,7,0,0x60,0x90,0x90,0x10,0x10,0x10,0x10,0x10,0x0,0x10,0x0,0x0},
{107,7,12,0,0,7,0,0x0,0x0,0x88,0x90,0xa0,0xc0,0xb8,0x80,0x80,0x80,0x0,0x0},
{108,7,12,0,0,7,0,0x0,0x0,0x30,0x20,0x20,0x20,0x20,0x20,0x20,0x60,0x0,0x0},
{109,7,12,0,0,7,0,0x0,0x0,0xa8,0xa8,0xa8,0xa8,0xa8,0xd0,0x0,0x0,0x0,0x0},
{110,7,12,0,0,7,0,0x0,0x0,0x88,0x88,0x88,0x88,0xd0,0xb0,0x0,0x0,0x0,0x0},
{111,7,12,0,0,7,0,0x0,0x0,0x70,0x88,0x88,0x88,0x88,0x70,0x0,0x0,0x0,0x0},
{112,7,12,0,0,7,0,0x80,0x80,0xb0,0xc8,0x88,0xc8,0xa8,0xb0,0x0,0x0,0x0,0x0},
{113,7,12,0,0,7,0,0x8,0x8,0x68,0x98,0x88,0x98,0xa8,0x60,0x0,0x0,0x0,0x0},
{114,7,12,0,0,7,0,0x0,0x0,0x80,0x80,0x80,0xc0,0xa8,0xb0,0x0,0x0,0x0,0x0},
{115,7,12,0,0,7,0,0x0,0x0,0xf0,0x88,0x10,0x60,0x88,0x70,0x0,0x0,0x0,0x0},
{116,7,12,0,0,7,0,0x0,0x0,0x70,0x50,0x40,0x40,0x40,0xf0,0x40,0x0,0x0,0x0},
{117,7,12,0,0,7,0,0x0,0x0,0x68,0x58,0x88,0x88,0x88,0x88,0x0,0x0,0x0,0x0},
{118,7,12,0,0,7,0,0x0,0x0,0x20,0x20,0x50,0x50,0x88,0x88,0x0,0x0,0x0,0x0},
{119,7,12,0,0,7,0,0x0,0x0,0x50,0xa8,0xa8,0x88,0x88,0x88,0x0,0x0,0x0,0x0},
{120,7,12,0,0,7,0,0x0,0x0,0x88,0x48,0x30,0x20,0x50,0x88,0x0,0x0,0x0,0x0},
{121,7,12,0,0,7,0,0x70,0x88,0x28,0x58,0x88,0x88,0x88,0x88,0x0,0x0,0x0,0x0},
{122,7,12,0,0,7,0,0x0,0x0,0xf8,0x40,0x20,0x10,0x18,0xf8,0x0,0x0,0x0,0x0},
{123,7,12,0,0,7,0,0x18,0x20,0x20,0x20,0x20,0xc0,0x20,0x20,0x20,0x18,0x0,0x0},
{124,7,12,0,0,7,0,0x0,0x8,0x8,0x8,0x8,0x8,0x0,0x8,0x8,0x8,0x0,0x0},
{125,7,12,0,0,7,0,0x40,0x20,0x20,0x20,0x20,0x18,0x20,0x20,0x20,0xc0,0x0,0x0},
{126,7,12,0,0,7,0,0x0,0x0,0x0,0x0,0x98,0xb4,0x64,0x0,0x0,0x0,0x0,0x0},
{127,7,12,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},
{255,0,0,0,0,0,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0}
};

GLubyte fwLetters7x14[][21] = {
{32,7,14,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},
{33,7,14,0,0,7,0,0x0,0x0,0x0,0x20,0x20,0x0,0x20,0x20,0x20,0x20,0x20,0x0,0x0,0x0},
{34,7,14,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x50,0x50,0x50,0x0,0x0},
{35,7,14,0,0,7,0,0x0,0x0,0x0,0x50,0x50,0xf8,0x50,0x50,0xf8,0x50,0x0,0x0,0x0,0x0},
{36,7,14,0,0,7,0,0x0,0x0,0x20,0x70,0xa8,0x28,0x30,0x60,0xa0,0xa8,0x70,0x20,0x0,0x0},
{37,7,14,0,0,7,0,0x0,0x0,0x0,0x98,0x54,0x54,0x2c,0x20,0x10,0xc8,0xa8,0xa0,0x60,0x0},
{38,7,14,0,0,7,0,0x0,0x0,0x0,0x74,0x88,0x8c,0x90,0x60,0x20,0x50,0x48,0x30,0x0,0x0},
{39,7,14,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20,0x10,0x10,0x0,0x0},
{40,7,14,0,0,7,0,0x0,0x0,0x0,0x20,0x40,0x40,0x80,0x80,0x80,0x80,0x80,0x40,0x40,0x20},
{41,7,14,0,0,7,0,0x0,0x0,0x0,0x20,0x10,0x10,0x8,0x8,0x8,0x8,0x10,0x10,0x20,0x0},
{42,7,14,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x20,0xa8,0x70,0x70,0xa8,0x20,0x0,0x0,0x0},
{43,7,14,0,0,7,0,0x0,0x0,0x0,0x0,0x20,0x20,0xf8,0x20,0x20,0x0,0x0,0x0,0x0,0x0},
{44,7,14,0,0,7,0,0x0,0x20,0x10,0x10,0x30,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},
{45,7,14,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x78,0x0,0x0,0x0,0x0,0x0,0x0},
{46,7,14,0,0,7,0,0x0,0x0,0x0,0x30,0x30,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},
{47,7,14,0,0,7,0,0x0,0x0,0x0,0x80,0x80,0x40,0x40,0x20,0x20,0x10,0x10,0x8,0x0,0x0},
{48,7,14,0,0,7,0,0x0,0x0,0x0,0x70,0x88,0x88,0xc8,0xb8,0x88,0x88,0x70,0x0,0x0,0x0},
{49,7,14,0,0,7,0,0x0,0x0,0x0,0x38,0x10,0x10,0x10,0x10,0x10,0x10,0x30,0x10,0x0,0x0},
{50,7,14,0,0,7,0,0x0,0x0,0x0,0xf8,0x80,0x40,0x20,0x10,0x8,0x88,0x88,0x70,0x0,0x0},
{51,7,14,0,0,7,0,0x0,0x0,0x0,0x70,0x88,0x8,0x8,0x30,0x8,0x88,0x70,0x0,0x0,0x0},
{52,7,14,0,0,7,0,0x0,0x0,0x0,0x10,0x10,0xf8,0x90,0x50,0x30,0x10,0x10,0x0,0x0,0x0},
{53,7,14,0,0,7,0,0x0,0x0,0x0,0x70,0x88,0x8,0x8,0x8,0xf0,0x80,0x80,0xf8,0x0,0x0},
{54,7,14,0,0,7,0,0x0,0x0,0x0,0x70,0x88,0x88,0x88,0xf0,0x80,0x80,0x90,0x70,0x0,0x0},
{55,7,14,0,0,7,0,0x0,0x0,0x0,0x20,0x20,0x20,0x20,0x20,0x10,0x8,0x8,0xf8,0x0,0x0},
{56,7,14,0,0,7,0,0x0,0x0,0x0,0x70,0x88,0x88,0x88,0x88,0x70,0x88,0x88,0x70,0x0,0x0},
{57,7,14,0,0,7,0,0x0,0x0,0x0,0x70,0x8,0x8,0x8,0x78,0x88,0x88,0x88,0x70,0x0,0x0},
{58,7,14,0,0,7,0,0x0,0x0,0x0,0x0,0x20,0x20,0x0,0x20,0x20,0x0,0x0,0x0,0x0,0x0},
{59,7,14,0,0,7,0,0x0,0x0,0x40,0x20,0x10,0x30,0x0,0x10,0x10,0x0,0x0,0x0,0x0,0x0},
{60,7,14,0,0,7,0,0x0,0x0,0x0,0x8,0x10,0x20,0x40,0x80,0x40,0x20,0x10,0x8,0x0,0x0},
{61,7,14,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x0,0x78,0x0,0x78,0x0,0x0,0x0,0x0,0x0},
{62,7,14,0,0,7,0,0x0,0x0,0x0,0x40,0x20,0x10,0x8,0x4,0x8,0x10,0x20,0x40,0x0,0x0},
{63,7,14,0,0,7,0,0x0,0x0,0x20,0x20,0x0,0x20,0x20,0x10,0x88,0x88,0x70,0x0,0x0,0x0},
{64,7,14,0,0,7,0,0x0,0x0,0x0,0x20,0x50,0x90,0xa8,0xa8,0xb8,0x88,0x48,0x30,0x0,0x0},
{65,7,14,0,0,7,0,0x0,0x0,0x0,0x88,0x88,0xf8,0x88,0x50,0x50,0x20,0x20,0x20,0x0,0x0},
{66,7,14,0,0,7,0,0x0,0x0,0x0,0xf0,0x88,0x88,0x88,0xf0,0x90,0x88,0x88,0xf0,0x0,0x0},
{67,7,14,0,0,7,0,0x0,0x0,0x0,0x70,0x88,0x80,0x80,0x80,0x80,0x88,0x88,0x70,0x0,0x0},
{68,7,14,0,0,7,0,0x0,0x0,0x0,0xf0,0x88,0x88,0x88,0x88,0x88,0x88,0xf0,0x0,0x0,0x0},
{69,7,14,0,0,7,0,0x0,0x0,0x0,0xf8,0x80,0x80,0x80,0xe0,0x80,0x80,0x80,0xf8,0x0,0x0},
{70,7,14,0,0,7,0,0x0,0x0,0x0,0x80,0x80,0x80,0x80,0xe0,0x80,0x80,0x80,0xf8,0x0,0x0},
{71,7,14,0,0,7,0,0x0,0x0,0x0,0x68,0x98,0x88,0x88,0x98,0x80,0x88,0x48,0x70,0x0,0x0},
{72,7,14,0,0,7,0,0x0,0x0,0x0,0x88,0x88,0x88,0x88,0xf8,0x88,0x88,0x88,0x0,0x0,0x0},
{73,7,14,0,0,7,0,0x0,0x0,0x0,0x70,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x70,0x0,0x0},
{74,7,14,0,0,7,0,0x0,0x0,0x0,0x60,0x90,0x10,0x10,0x10,0x10,0x10,0x10,0x30,0x0,0x0},
{75,7,14,0,0,7,0,0x0,0x0,0x0,0x8c,0x90,0xa0,0xc0,0xa0,0x90,0x88,0x88,0x0,0x0,0x0},
{76,7,14,0,0,7,0,0x0,0x0,0x0,0xf8,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x0,0x0,0x0},
{77,7,14,0,0,7,0,0x0,0x0,0x0,0x88,0x88,0x88,0xa8,0xa8,0xd8,0x88,0x88,0x0,0x0,0x0},
{78,7,14,0,0,7,0,0x0,0x0,0x0,0x88,0x98,0xb8,0xa8,0xc8,0xc8,0x88,0x88,0x0,0x0,0x0},
{79,7,14,0,0,7,0,0x0,0x0,0x0,0x70,0x88,0x88,0x88,0x88,0x88,0x88,0x70,0x0,0x0,0x0},
{80,7,14,0,0,7,0,0x0,0x0,0x0,0x80,0x80,0x80,0x80,0x80,0xf0,0x88,0x88,0xf0,0x0,0x0},
{81,7,14,0,0,7,0,0x0,0x18,0x20,0x70,0xc8,0x88,0x88,0x88,0x88,0x88,0x88,0x70,0x0,0x0},
{82,7,14,0,0,7,0,0x0,0x0,0x0,0x88,0x90,0x90,0xa0,0xf0,0x88,0x88,0xf0,0x0,0x0,0x0},
{83,7,14,0,0,7,0,0x0,0x0,0x0,0x70,0x88,0x8,0x8,0x10,0x60,0x80,0x88,0x70,0x0,0x0},
{84,7,14,0,0,7,0,0x0,0x0,0x0,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0xf8,0x0,0x0,0x0},
{85,7,14,0,0,7,0,0x0,0x0,0x0,0x70,0x88,0x88,0x88,0x88,0x88,0x88,0x88,0x0,0x0,0x0},
{86,7,14,0,0,7,0,0x0,0x0,0x0,0x20,0x20,0x20,0x50,0x50,0x88,0x88,0x88,0x0,0x0,0x0},
{87,7,14,0,0,7,0,0x0,0x0,0x0,0x88,0x88,0xd8,0xa8,0xa8,0x88,0x88,0x88,0x0,0x0,0x0},
{88,7,14,0,0,7,0,0x0,0x0,0x0,0x88,0x88,0x50,0x50,0x20,0x50,0x50,0x88,0x88,0x0,0x0},
{89,7,14,0,0,7,0,0x0,0x0,0x0,0x20,0x20,0x20,0x20,0x50,0x50,0x50,0x88,0x88,0x0,0x0},
{90,7,14,0,0,7,0,0x0,0x0,0x0,0xf8,0x80,0x40,0x40,0x20,0x10,0x10,0x8,0xf8,0x0,0x0},
{91,7,14,0,0,7,0,0x0,0x0,0x30,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x30,0x0,0x0},
{92,7,14,0,0,7,0,0x0,0x0,0x0,0x8,0x8,0x10,0x10,0x20,0x20,0x40,0x40,0x80,0x80,0x0},
{93,7,14,0,0,7,0,0x0,0x0,0x30,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x30,0x0,0x0},
{94,7,14,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x88,0x50,0x20,0x0,0x0},
{95,7,14,0,0,7,0,0x0,0xfc,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},
{96,7,14,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10,0x20,0x20,0x0,0x0},
{97,7,14,0,0,7,0,0x0,0x0,0x0,0x68,0x98,0x88,0x78,0x8,0x88,0x70,0x0,0x0,0x0,0x0},
{98,7,14,0,0,7,0,0x0,0x0,0x0,0xb0,0xc8,0x88,0x88,0x88,0xc8,0xb0,0x80,0x80,0x0,0x0},
{99,7,14,0,0,7,0,0x0,0x0,0x0,0x70,0x88,0x80,0x80,0x80,0x88,0x70,0x0,0x0,0x0,0x0},
{100,7,14,0,0,7,0,0x0,0x0,0x0,0x68,0x98,0x88,0x88,0x98,0x68,0x8,0x8,0x8,0x0,0x0},
{101,7,14,0,0,7,0,0x0,0x0,0x0,0x70,0x88,0x80,0xb8,0xc8,0x88,0x70,0x0,0x0,0x0,0x0},
{102,7,14,0,0,7,0,0x0,0x0,0x0,0x20,0x20,0x20,0x20,0x20,0x70,0x20,0x24,0x18,0x0,0x0},
{103,7,14,0,0,7,0,0x78,0x88,0x8,0x68,0x98,0x88,0x88,0x88,0x98,0x68,0x0,0x0,0x0,0x0},
{104,7,14,0,0,7,0,0x0,0x0,0x0,0x88,0x88,0x88,0x88,0x88,0xc8,0xb0,0x80,0x80,0x0,0x0},
{105,7,14,0,0,7,0,0x0,0x0,0x0,0x10,0x10,0x10,0x10,0x10,0x10,0x30,0x0,0x10,0x0,0x0},
{106,7,14,0,0,7,0,0x60,0x90,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x30,0x0,0x10,0x0,0x0},
{107,7,14,0,0,7,0,0x0,0x0,0x0,0x84,0x88,0x90,0xe0,0xa0,0x90,0x90,0x80,0x80,0x0,0x0},
{108,7,14,0,0,7,0,0x0,0x0,0x0,0x30,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x60,0x0,0x0},
{109,7,14,0,0,7,0,0x0,0x0,0x0,0xa8,0xa8,0xa8,0xa8,0xa8,0xe8,0x90,0x0,0x0,0x0,0x0},
{110,7,14,0,0,7,0,0x0,0x0,0x0,0x88,0x88,0x88,0x88,0x88,0xc8,0xb0,0x0,0x0,0x0,0x0},
{111,7,14,0,0,7,0,0x0,0x0,0x0,0x70,0x88,0x88,0x88,0x88,0x88,0x70,0x0,0x0,0x0,0x0},
{112,7,14,0,0,7,0,0x80,0x80,0x80,0xb0,0xc8,0xc8,0x88,0x88,0xc8,0xb0,0x0,0x0,0x0,0x0},
{113,7,14,0,0,7,0,0xc,0x8,0x8,0x68,0x98,0x88,0x88,0x88,0x78,0x0,0x0,0x0,0x0,0x0},
{114,7,14,0,0,7,0,0x0,0x0,0x0,0x80,0x80,0x80,0x80,0x80,0xc8,0xb0,0x0,0x0,0x0,0x0},
{115,7,14,0,0,7,0,0x0,0x0,0x0,0x70,0x88,0x8,0x70,0x80,0x80,0x78,0x0,0x0,0x0,0x0},
{116,7,14,0,0,7,0,0x0,0x0,0x0,0x10,0x28,0x20,0x20,0x20,0x70,0x20,0x20,0x0,0x0,0x0},
{117,7,14,0,0,7,0,0x0,0x0,0x0,0x6c,0x98,0x88,0x88,0x88,0x88,0x0,0x0,0x0,0x0,0x0},
{118,7,14,0,0,7,0,0x0,0x0,0x0,0x20,0x20,0x50,0x50,0x50,0x88,0x88,0x0,0x0,0x0,0x0},
{119,7,14,0,0,7,0,0x0,0x0,0x0,0x50,0x50,0xa8,0xa8,0xa8,0x88,0x88,0x0,0x0,0x0,0x0},
{120,7,14,0,0,7,0,0x0,0x0,0x0,0x88,0x88,0x50,0x20,0x50,0x88,0x88,0x0,0x0,0x0,0x0},
{121,7,14,0,0,7,0,0x30,0xc8,0x8,0x68,0x98,0x88,0x88,0x88,0x88,0x88,0x0,0x0,0x0,0x0},
{122,7,14,0,0,7,0,0x0,0x0,0x0,0xf8,0x80,0x40,0x20,0x10,0x8,0xf8,0x0,0x0,0x0,0x0},
{123,7,14,0,0,7,0,0x0,0x0,0x10,0x20,0x20,0x20,0x20,0xc0,0x20,0x20,0x20,0x20,0x10,0x0},
{124,7,14,0,0,7,0,0x0,0x0,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x0,0x0},
{125,7,14,0,0,7,0,0x0,0x0,0x80,0x40,0x40,0x40,0x40,0x20,0x40,0x40,0x40,0x40,0x80,0x0},
{126,7,14,0,0,7,0,0x0,0x0,0x0,0x0,0x0,0x0,0x10,0xa8,0xa8,0x40,0x0,0x0,0x0,0x0},
{255,0,0,0,0,0,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0}
};

GLubyte fwLetters8x15[][22] = {
{32,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},
{33,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x20,0x20,0x0,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x0},
{35,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x24,0x24,0x24,0xfe,0x24,0x24,0x24,0xfe,0x24,0x0,0x0},
{36,8,15,0,0,8,0,0x0,0x0,0x0,0x10,0x38,0x54,0x94,0x14,0x18,0x10,0x70,0x90,0x94,0x78,0x10},
{37,8,15,0,0,8,0,0x0,0x0,0x0,0x80,0x44,0x4a,0x2a,0x34,0x10,0x10,0x48,0xa8,0xa4,0x44,0x0},
{38,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x74,0x88,0x94,0xa0,0x40,0x40,0xa0,0x90,0x50,0x20,0x0},
{39,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x20,0x20,0x30,0x30,0x0,0x0},
{40,8,15,0,0,8,0,0x0,0x0,0x0,0x8,0x10,0x20,0x20,0x40,0x40,0x40,0x40,0x20,0x20,0x10,0x8},
{41,8,15,0,0,8,0,0x0,0x0,0x40,0x20,0x10,0x10,0x8,0x8,0x8,0x8,0x8,0x10,0x10,0x20,0x40},
{42,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x0,0x0,0x10,0x54,0x38,0x38,0x54,0x10,0x0,0x0,0x0},
{43,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x10,0x10,0x10,0xfe,0x10,0x10,0x10,0x10,0x0,0x0,0x0},
{44,8,15,0,0,8,0,0x0,0x0,0x20,0x10,0x18,0x18,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},
{45,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf8,0x0,0x0,0x0,0x0,0x0,0x0},
{46,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x30,0x30,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},
{47,8,15,0,0,8,0,0x0,0x0,0x40,0x40,0x20,0x20,0x10,0x10,0x8,0x8,0x4,0x4,0x2,0x2,0x0},
{48,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x78,0x84,0x84,0xc4,0xa4,0x9c,0x84,0x84,0x84,0x78,0x0},
{49,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x38,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x30,0x10,0x0},
{50,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0xfe,0x80,0x40,0x20,0x10,0x8,0x6,0x82,0x82,0x7c,0x0},
{51,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x78,0x84,0x4,0x4,0x4,0x18,0x4,0x4,0x84,0x78,0x0},
{52,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x8,0x8,0x8,0x8,0xfc,0x88,0x48,0x28,0x18,0x8,0x0},
{53,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x78,0x84,0x4,0x4,0x84,0xf8,0x80,0x80,0x80,0xfc,0x0},
{54,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x7c,0x84,0x82,0xc2,0xa4,0x98,0x80,0x84,0x44,0x38,0x0},
{55,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x20,0x20,0x10,0x10,0x10,0x10,0x8,0x4,0x4,0xfc,0x0},
{56,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x78,0x84,0x84,0x84,0x84,0x84,0x78,0x84,0x84,0x78,0x0},
{57,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x78,0x84,0x4,0x34,0x4c,0x84,0x84,0x84,0x44,0x38,0x0},
{58,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x30,0x30,0x0,0x0,0x0,0x30,0x30,0x0,0x0,0x0,0x0},
{59,8,15,0,0,8,0,0x0,0x40,0x20,0x10,0x30,0x30,0x0,0x0,0x30,0x30,0x0,0x0,0x0,0x0,0x0},
{60,8,15,0,0,8,0,0x0,0x0,0x0,0x4,0x8,0x10,0x20,0x40,0x80,0x40,0x20,0x10,0x8,0x4,0x0},
{61,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf8,0x0,0x0,0xf8,0x0,0x0,0x0,0x0},
{62,8,15,0,0,8,0,0x0,0x0,0x80,0x40,0x20,0x10,0x8,0x4,0x4,0x8,0x10,0x20,0x40,0x80,0x0},
{63,8,15,0,0,8,0,0x0,0x0,0x0,0x10,0x10,0x0,0x0,0x10,0x18,0x4,0x2,0x82,0x44,0x38,0x0},
{64,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x38,0x44,0x80,0x98,0xa4,0xa4,0x9c,0x84,0x48,0x30,0x0},
{65,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x84,0x84,0xfc,0x84,0x48,0x48,0x48,0x30,0x30,0x0,0x0},
{66,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0xf8,0x84,0x84,0x84,0x84,0xf8,0x84,0x84,0x84,0xf8,0x0},
{67,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x78,0x84,0x80,0x80,0x80,0x80,0x80,0x80,0x84,0x78,0x0},
{68,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0xf0,0x88,0x84,0x84,0x84,0x84,0x84,0x88,0xf0,0x0,0x0},
{69,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0xfc,0x80,0x80,0x80,0x80,0xf0,0x80,0x80,0xfc,0x0,0x0},
{70,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x80,0x80,0x80,0x80,0x80,0xf0,0x80,0x80,0x80,0xfe,0x0},
{71,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x7a,0x86,0x82,0x82,0x82,0x8c,0x80,0x80,0x44,0x38,0x0},
{72,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x84,0x84,0x84,0x84,0xfc,0x84,0x84,0x84,0x84,0x84,0x0},
{73,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x38,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x38,0x0},
{74,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x70,0x88,0x88,0x8,0x8,0x8,0x8,0x8,0x8,0x18,0x0},
{75,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x86,0x88,0x90,0xa0,0xc0,0xa0,0x90,0x88,0x84,0x80,0x0},
{76,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0xfc,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x0},
{77,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x82,0x82,0x92,0x92,0xaa,0xaa,0xc6,0xc6,0x82,0x82,0x0},
{78,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x84,0x8c,0x8c,0x94,0x94,0xa4,0xa4,0xc4,0xc4,0x84,0x0},
{79,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x78,0x84,0x84,0x84,0x84,0x84,0x84,0x84,0x84,0x78,0x0},
{80,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x80,0x80,0x80,0x80,0xb8,0xc4,0x84,0x84,0x84,0xf8,0x0},
{81,8,15,0,0,8,0,0x0,0x4,0x18,0x20,0x7c,0xa2,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x7c,0x0},
{82,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x82,0x84,0x8c,0x88,0xfc,0x82,0x82,0x82,0x82,0xfc,0x0},
{83,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x78,0x84,0x4,0x4,0x18,0x60,0x80,0x80,0x84,0x7c,0x0},
{84,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0xfe,0x0},
{85,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x7c,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x0},
{86,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x10,0x10,0x28,0x44,0x44,0x44,0x44,0x82,0x82,0x82,0x0},
{87,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x44,0x44,0xaa,0xaa,0x92,0x92,0x92,0x82,0x82,0x0,0x0},
{88,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x84,0x84,0x48,0x48,0x30,0x30,0x4c,0x44,0x84,0x84,0x0},
{89,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x10,0x10,0x10,0x10,0x10,0x28,0x28,0x44,0x82,0x82,0x0},
{90,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0xfe,0x80,0x40,0x40,0x20,0x10,0x8,0x4,0x4,0xfe,0x0},
{91,8,15,0,0,8,0,0x0,0x0,0x0,0xe0,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0xe0,0x0},
{92,8,15,0,0,8,0,0x0,0x0,0x4,0x4,0x8,0x8,0x10,0x10,0x20,0x20,0x40,0x40,0x80,0x80,0x0},
{93,8,15,0,0,8,0,0x0,0x0,0x0,0x38,0x8,0x8,0x8,0x8,0x8,0x8,0x8,0x8,0x8,0x38,0x0},
{94,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x44,0x28,0x10,0x0},
{95,8,15,0,0,8,0,0x0,0xfe,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},
{96,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20,0x40,0xc0,0xc0,0x0},
{97,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x74,0x88,0x98,0x68,0x8,0x88,0x70,0x0,0x0,0x0,0x0},
{98,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0xb8,0xc4,0x84,0xc4,0xc4,0xb8,0x80,0x80,0x80,0x0,0x0},
{99,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x78,0x84,0x80,0x80,0x80,0x84,0x78,0x0,0x0,0x0,0x0},
{100,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x74,0x8c,0x8c,0x84,0x8c,0x74,0x4,0x4,0x4,0x0,0x0},
{101,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x78,0x84,0x80,0xbc,0xc4,0x84,0x78,0x0,0x0,0x0,0x0},
{102,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x20,0x20,0x20,0x20,0x78,0x20,0x20,0x24,0x3c,0x0,0x0},
{103,8,15,0,0,8,0,0x18,0x64,0x4,0x4,0x34,0x4c,0x84,0x84,0x84,0x8c,0x74,0x0,0x0,0x0,0x0},
{104,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x84,0x84,0x84,0x84,0x84,0xc4,0xb8,0x80,0x80,0x80,0x0},
{105,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x10,0x10,0x10,0x10,0x10,0x10,0x30,0x0,0x10,0x0,0x0},
{106,8,15,0,0,8,0,0x40,0xa0,0x90,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x30,0x0,0x10,0x0},
{107,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x84,0x98,0xb0,0xc0,0xa0,0x90,0x88,0x80,0x80,0x0,0x0},
{108,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x18,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x30,0x0,0x0},
{109,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x54,0x54,0x54,0x54,0x54,0x54,0xa8,0x0,0x0,0x0,0x0},
{110,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x84,0x84,0x84,0x84,0x84,0xc8,0xb8,0x0,0x0,0x0,0x0},
{111,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x78,0x84,0x84,0x84,0x84,0x84,0x78,0x0,0x0,0x0,0x0},
{112,8,15,0,0,8,0,0x80,0x80,0x80,0x80,0xb8,0xa4,0xc4,0x84,0x84,0xc4,0xa4,0x18,0x0,0x0,0x0},
{113,8,15,0,0,8,0,0x2,0x4,0x4,0x4,0x74,0x8c,0x8c,0x84,0x84,0x8c,0x74,0x0,0x0,0x0,0x0},
{114,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x80,0x80,0x80,0x80,0xc0,0xa4,0xb8,0x0,0x0,0x0,0x0},
{115,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0xf8,0x84,0x4,0x38,0x40,0x84,0x78,0x0,0x0,0x0,0x0},
{116,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x30,0x28,0x20,0x20,0x20,0x20,0x78,0x20,0x20,0x0,0x0},
{117,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x74,0x4c,0x84,0x84,0x84,0x84,0x84,0x0,0x0,0x0,0x0},
{118,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x30,0x30,0x48,0x48,0x84,0x84,0x84,0x0,0x0,0x0,0x0},
{119,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x24,0x5a,0x92,0x92,0x82,0x82,0x82,0x0,0x0,0x0,0x0},
{120,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x84,0x84,0x48,0x30,0x48,0x84,0x84,0x0,0x0,0x0,0x0},
{121,8,15,0,0,8,0,0x38,0x44,0x84,0x4,0x74,0x8c,0x84,0x84,0x84,0x84,0x0,0x0,0x0,0x0,0x0},
{122,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0xfc,0x80,0x40,0x20,0x10,0x8,0xfc,0x0,0x0,0x0,0x0},
{123,8,15,0,0,8,0,0x0,0x0,0x30,0x40,0x40,0x40,0x40,0x40,0xc0,0x40,0x40,0x40,0x40,0x30,0x0},
{124,8,15,0,0,8,0,0x0,0x0,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x0},
{125,8,15,0,0,8,0,0x0,0x0,0x0,0x60,0x10,0x10,0x10,0x10,0x18,0x10,0x10,0x10,0x10,0x60,0x0},
{126,8,15,0,0,8,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x98,0xb4,0x64,0x0,0x0,0x0,0x0,0x0},
{255,0,0,0,0,0,0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0}
};
/* <<< bitmap fonts */

typedef struct {int x; int y;} XY;
//GLuint fwFontOffset[3];
//XY fwFontSize[3];
#include <list.h>
//static int showConText = 0;
//s_list_t *conlist;
//int concount;

//static int loopcount = 0;
////extern int clipPlane;
//static int hadString = 0;
//
//int showButtons =0;
#define mbuts 12
//int nbuts = mbuts;
//int butrect[4][mbuts];
//int butStatus[mbuts];
//textureTableIndexStruct_s butts[mbuts][2];
char * butFnames[mbuts] = {"walk.png","fly.png","examine.png","level.png","headlight.png","collision.png","prev.png","next.png","help.png","messages.png","options.png","reload.png"};//"flyEx.png",
//GLubyte *bmIcon[mbuts]; 
//int butsLoaded = 0;
//int isOver = -1;
//int iconSize = 32;
//int buttonType = 1; /* 0 = rgba .png 1= .c bitmap (see above) */
//int savePng2dotc = 0; /* if you read png and want to save to a bitmap .c struct, put 1 */
////bool texture_load_from_file(textureTableIndexStruct_s* this_tex, char *filename);

#define STATUS_LEN 2000

typedef struct pstatusbar{
	int loopcount;// = 0;
	int hadString;// = 0;
	int initDone;
	int showButtons;// =0;
	int nbuts;// = mbuts;
	int butrect[4][mbuts];
	int butStatus[mbuts];
	textureTableIndexStruct_s butts[mbuts][2];
	GLubyte *bmIcon[mbuts]; 
	int butsLoaded;// = 0;
	int isOver;// = -1;
	int iconSize;// = 32;
	int buttonType;// = 1; /* 0 = rgba .png 1= .c bitmap (see above) */
	int savePng2dotc;// = 0; /* if you read png and want to save to a bitmap .c struct, put 1 */
	int showConText;// = 0;
	s_list_t *conlist;
	int concount;
	int fontInitialized;// = 0;
	GLuint fwFontOffset[3];
	XY fwFontSize[3];
	int sb_hasString;// = FALSE;
	struct Uni_String *myline;
	char buffer[200];
	char messagebar[200];
	int bmfontsize;// = 2; /* 0,1 or 2 */
	int optionsLoaded;// = 0;
	char * optionsVal[15];
	int osystem;// = 3; //mac 1btn = 0, mac nbutton = 1, linux game descent = 2, windows =3
	int showOptions;// =0;
	XY bmWH;// = {10,15}; /* simple bitmap font from redbook above, width and height in pixels */
	int posType; //1 == glRasterPos (opengl < 1.4), 0= glWindowPos (opengl 1.4+)
}* ppstatusbar;
void *statusbar_constructor(){
	void *v = malloc(sizeof(struct pstatusbar));
	memset(v,0,sizeof(struct pstatusbar));
	return v;
}
void statusbar_init(struct tstatusbar *t){
	//public
	//private
	t->prv = statusbar_constructor();
	{
		ppstatusbar p = (ppstatusbar)t->prv;
		p->loopcount = 0;
		p->hadString = 0;

		p->showButtons =0;
		p->nbuts = mbuts;// -1; // leave Reload out of pre2
		//p->butrect[4][mbuts];
		//p->butStatus[mbuts];
		//p->butts[mbuts][2];
		//p->bmIcon[mbuts]; 
		p->butsLoaded = 0;
		p->isOver = -1;
		p->iconSize = 32;
		p->buttonType = 1; /* 0 = rgba .png 1= .c bitmap (see above) (put 0 to read .png, write C) */
		p->savePng2dotc = 0; /* if you read png and want to save to a bitmap .c struct, put 1 */
		p->showConText = 0;
		//p->conlist;
		//p->concount;
		p->fontInitialized = 0;
		//p->fwFontOffset[3];
		//p->fwFontSize[3];
		p->sb_hasString = FALSE;
		p->initDone = FALSE;
		//p->myline;
		//p->buffer[200];
		//p->messagebar[200];
		p->bmfontsize = 2; /* 0,1 or 2 */
		p->optionsLoaded = 0;
		//p->optionsVal[15];
		p->osystem = 3; //mac 1btn = 0, mac nbutton = 1, linux game descent = 2, windows =3
		p->showOptions =0;
		p->bmWH.x = 10;
		p->bmWH.y = 15; //{10,15}; /* simple bitmap font from redbook above, width and height in pixels */
		p->posType = 0; //assume ogl 1.4+, and correct if not
	}
}
//ppstatusbar p = (ppstatusbar)gglobal()->statusbar.prv;

void fwMakeRasterFonts()
{
	int ichar;
	GLuint i;
	GLuint fwFontOffset7x12;
	GLuint fwFontOffset7x14;
	GLuint fwFontOffset8x15;
	ppstatusbar p = (ppstatusbar)gglobal()->statusbar.prv;

   FW_GL_PIXELSTOREI(GL_UNPACK_ALIGNMENT, 1);
   fwFontOffset7x12 = glGenLists (128);
   for(i=0;i<128;i++)
   {
		ichar = fwLetters7x12[i][0];
		if( ichar == 255 )break;
	   FW_GL_NEWLIST(fwFontOffset7x12 + fwLetters7x12[i][0],GL_COMPILE);
	   FW_GL_BITMAP((GLsizei)fwLetters7x12[i][1],(GLsizei)fwLetters7x12[i][2],
		   (GLfloat)fwLetters7x12[i][3],(GLfloat)fwLetters7x12[i][4],
		   (GLfloat)fwLetters7x12[i][5], (GLfloat)fwLetters7x12[i][6],
		   &fwLetters7x12[i][7]);
	   FW_GL_ENDLIST();
   }
   fwFontOffset7x14 = glGenLists (128);
   for(i=0;i<128;i++)
   {
		ichar = fwLetters7x14[i][0];
		if( ichar == 255 )break;
	   FW_GL_NEWLIST(fwFontOffset7x14 + fwLetters7x14[i][0],GL_COMPILE);
	   FW_GL_BITMAP((GLsizei)fwLetters7x14[i][1],(GLsizei)fwLetters7x14[i][2],
		   (GLfloat)fwLetters7x14[i][3],(GLfloat)fwLetters7x14[i][4],
		   (GLfloat)fwLetters7x14[i][5], (GLfloat)fwLetters7x14[i][6],
		   &fwLetters7x14[i][7]);
	   FW_GL_ENDLIST();
   }
   fwFontOffset8x15 = glGenLists (128);
   for(i=0;i<128;i++)
   {
		ichar = fwLetters8x15[i][0];
		if( ichar == 255 )break;
	   FW_GL_NEWLIST(fwFontOffset8x15 + fwLetters8x15[i][0],GL_COMPILE);
	   FW_GL_BITMAP((GLsizei)fwLetters8x15[i][1],(GLsizei)fwLetters8x15[i][2],
		   (GLfloat)fwLetters8x15[i][3],(GLfloat)fwLetters8x15[i][4],
		   (GLfloat)fwLetters8x15[i][5], (GLfloat)fwLetters8x15[i][6],
		   &fwLetters8x15[i][7]);
	   FW_GL_ENDLIST();
   }
   p->fwFontOffset[0] = fwFontOffset7x12;
   p->fwFontOffset[1] = fwFontOffset7x14;
   p->fwFontOffset[2] = fwFontOffset8x15;
   p->fwFontSize[0].x = 7;
   p->fwFontSize[0].y = 12;
   p->fwFontSize[1].x = 7;
   p->fwFontSize[1].y = 14;
   p->fwFontSize[2].x = 8;
   p->fwFontSize[2].y = 15;

}

//static int fontInitialized = 0;
//XY bmWH = {10,15}; /* simple bitmap font from redbook above, width and height in pixels */

void initFont(void)
{
	/*initialize raster bitmap font above */
  // FW_GL_SHADEMODEL (GL_FLAT);
	ppstatusbar p = (ppstatusbar)gglobal()->statusbar.prv;

   fwMakeRasterFonts();
   p->fontInitialized = 1;
}
//int bmfontsize = 2; /* 0,1 or 2 */
void printString(char *s)
{
	ppstatusbar p = (ppstatusbar)gglobal()->statusbar.prv;

   FW_GL_PUSH_ATTRIB(GL_LIST_BIT);
   FW_GL_LISTBASE(p->fwFontOffset[p->bmfontsize]);
   p->bmWH.x = p->fwFontSize[p->bmfontsize].x;
   p->bmWH.y = p->fwFontSize[p->bmfontsize].y;
   FW_GL_CALLLISTS((int) strlen(s), GL_UNSIGNED_BYTE, (GLubyte *) s);
   FW_GL_POP_ATTRIB();
}

//static int sb_hasString = FALSE;
//static struct Uni_String *myline;
//
//static char buffer[200];
//
//#define STATUS_LEN 2000
//char messagebar[200];
void render_init(void);

/* make sure that on a re-load that we re-init */
void kill_status (void) {
	/* hopefully, by this time, rendering has been stopped */
	ppstatusbar p = (ppstatusbar)gglobal()->statusbar.prv;

	p->sb_hasString = FALSE;
}


/* trigger a update */
void update_status(char* msg) {
	ppstatusbar p = (ppstatusbar)gglobal()->statusbar.prv;

	if (msg==NULL) p->sb_hasString = FALSE;
	else {
		p->sb_hasString = TRUE;
		strcpy (p->buffer,msg);
	}
}


/* start cheapskate widgets >>>> */
int lenOptions = 15;
char * optionsText[15] = {
" EAI",
"Stereovision",
" shutter",
" sidebyside",
" anaglyph",
"Anaglyph",
" RGBACM",
"        left",
"        right",
"Eyebase",
"<       >",
"ScreenDistance",
"<       >",
"Toe In",
"<       >"
};
//int optionsLoaded = 0;
//char * optionsVal[15];
void setOptionsVal()
{
}
static char * RGBACM = "RGBACM";
void initOptionsVal()
{
	int i,j;
	X3D_Viewer *viewer;
	ppstatusbar p = (ppstatusbar)gglobal()->statusbar.prv;
	viewer = Viewer();

	for(i=0;i<lenOptions;i++)
	{
		p->optionsVal[i] = (char*)malloc(8);
		for(j=0;j<8;j++) p->optionsVal[i][j] = ' ';
		p->optionsVal[i][7] = '\0';
	}
	if(fwl_getp_eai()) p->optionsVal[0][0] = '*';
	if(viewer->shutterGlasses) p->optionsVal[2][0] = '*';
	if(viewer->sidebyside) p->optionsVal[3][0] = '*';
	if(viewer->anaglyph) p->optionsVal[4][0] = '*';
	// Q. what / which can I rely on for the anaglyph color or index?
	// A. Viewer.iprog[iside={0,1}] = 0-5 for RGBACM  
	p->optionsVal[7][viewer->iprog[0]+1] = RGBACM[viewer->iprog[0]]; //'R';
	p->optionsVal[8][viewer->iprog[1]+1] = RGBACM[viewer->iprog[1]]; //'G';
	sprintf(p->optionsVal[10],"  %4.3f",viewer->eyedist); //.eyebase); //.060f);
	sprintf(p->optionsVal[12],"  %4.3f",viewer->screendist); //.6f);
	sprintf(p->optionsVal[14],"  %4.3f",viewer->stereoParameter); //.toein.4f);

	p->optionsLoaded = 1;
}
void updateOptionsVal()
{
	/* on each loop we refresh the hud Options state from the viewer state, 
	in case others - via keyboard or gui - are also updating the viewer state*/
	initOptionsVal();
}
/* the optionsCase char is used in a switch case later to involk the appropriate function */
char * optionsCase[15] = {
"TTTT",
"            ",
"11111111",
"22222222222",
"333333334",
"        ",
"       ",
" rgbacm     ",
" RGBACM      ",
"       ",
"556666677",
"              ",
"DDEEEEEFF",
"      ",
"HHIIIIIJJ"
};

XY mouse2screen(int x, int y)
{
	XY xy;
	xy.x = x;
	xy.y = gglobal()->display.screenHeight -y;
	return xy;
}
XY screen2text(int x, int y)
{
	XY rc;
	ppstatusbar p; 
	ttglobal tg = gglobal();
	p = (ppstatusbar)tg->statusbar.prv;

	rc.x = x/p->bmWH.x; //10; 
	rc.y = (int)((tg->display.screenHeight -y)/p->bmWH.y); //15.0 ); 
	return rc;
}
XY text2screen( int col, int row)
{
	XY xy;
	ppstatusbar p; 
	ttglobal tg = gglobal();
	p = (ppstatusbar)tg->statusbar.prv;

	xy.x = col*p->bmWH.x; //10; 
	xy.y = tg->display.screenHeight - (row+1)*p->bmWH.y; //15;
	return xy;
}
void printTextCursor()
{
	XY sxy,txy,cxy;
	float rgba[4];
	ppstatusbar p; 
	ttglobal tg = gglobal();
	p = (ppstatusbar)tg->statusbar.prv;

	sxy = mouse2screen(tg->Mainloop.currentX[0],tg->Mainloop.currentY[0]);
	txy = screen2text(sxy.x,sxy.y);
	cxy = text2screen(txy.x,txy.y);
	rgba[0] = .5f; rgba[1] = .5f; rgba[2] = .5f, rgba[3] = .75f;
	if(p->posType==1)
		FW_GL_RASTERPOS2I(cxy.x,cxy.y)
	else
		FW_GL_WINDOWPOS2I(cxy.x,cxy.y)
	FW_GL_PIXELZOOM(p->bmWH.x,p->bmWH.y); 
	FW_GL_DRAWPIXELS(1,1,GL_RGBA,GL_FLOAT,rgba);
	//restore
	FW_GL_PIXELZOOM(1.0f,1.0f);
}

void printOptions()
{
	int j; 
	ppstatusbar p = (ppstatusbar)gglobal()->statusbar.prv;

	printTextCursor();
	if(!p->optionsLoaded) initOptionsVal();
	updateOptionsVal(); //ideally we would be stateless in the hud, let Viewer hold state, that way other gui/shortcuts can be used

	for(j=0;j<lenOptions;j++)
	{
		XY xy = text2screen(0,j);
		if(p->posType==1)
			FW_GL_RASTERPOS2I(xy.x,xy.y)
		else
			FW_GL_WINDOWPOS2I(xy.x,xy.y)
		printString(p->optionsVal[j]);  /* "  0.050  " */
		if(p->posType==1)
			FW_GL_RASTERPOS2I(xy.x,xy.y)
		else
			FW_GL_WINDOWPOS2I(xy.x,xy.y)
		printString(optionsText[j]); /* "<       >" */
	}
}

//extern int showOptions;
//int showOptions =0;
void handleOptionPress()
{
	/* general idea: we don't update the hud/option state here - just the Viewer state - then 
	  refresh the hud/options state from the Viewer on each statusbar draw iteration
	*/
	int opt;
	XY xys;
	XY xyt;
	X3D_Viewer *viewer;
	ttglobal tg = gglobal();
	viewer = Viewer();

	xys = mouse2screen(tg->Mainloop.currentX[0],tg->Mainloop.currentY[0]);
	xyt = screen2text(xys.x,xys.y);
	opt = ' ';
	if( 0 <= xyt.y && xyt.y < lenOptions )
	{
		int len = (int) strlen(optionsCase[xyt.y]);
		if( xyt.x < len )
		{
			/* we are on an options line */
			opt = optionsCase[xyt.y][xyt.x];
		}
	}

	/* we're clicking a sensitive area.  */
	switch(opt) 
	{
	case 'T': {
		/* EAI */
		/* Note, this is actually useless (I suspect) because the EAI would already have started / (or ignored) */
		printf("toggle EAI");
		/* fwl_setp_eai(1 - fwl_getp_eai()); */
		break;}
	case '1': 
	case '2': 
	case '3': 
		toggleOrSetStereo(opt-'0');
		break;
	case 'r': 
	case 'g': 
	case 'b': 
	case 'a': 
	case 'c': 
	case 'm': 
		setAnaglyphSideColor(toupper(opt),0);
		break;
	case 'R': 
	case 'G': 
	case 'B': 
	case 'A': 
	case 'C': 
	case 'M': 
		setAnaglyphSideColor(opt,1);
		break;
	case '5': {
		/* eyebase */
		printf("reduce eyebase");
		viewer->eyedist *= .9;
		updateEyehalf();
		break;}
	case '6': {
		/* eyebase */
		printf("set eyebase");
	    //fwl_set_EyeDist(optarg);
		break;}
	case '7': {
		/* eyebase */
		printf("increase eyebase");
		viewer->eyedist *= 1.1;
		updateEyehalf();
		break;}
	case 'D': {
		/* screendist */
		printf("reduce screendist");
		viewer->screendist -= .02; //*= .9;
		updateEyehalf();
		break;}
	case 'E': {
		/* screendist */
		printf("set screendist");
		break;}
	case 'F': {
		/* screendist */
		printf("increase screendist");
		viewer->screendist += .02; //*= 1.1;
		updateEyehalf();
		break;}
	case 'H': {
		/* toein */
		printf("reduce toe-in");
		viewer->stereoParameter *= .9;
		updateEyehalf();
		break;}
	case 'I': {
		/* toein */
		printf("set toe-in");
		break;}
	case 'J': {
		/* toein */
		printf("increase toe-in");
		viewer->stereoParameter *= 1.1;
		updateEyehalf();
		break;}
	default: {break;}
	}
}
/* <<< end cheapskate widgets */



//int osystem = 3; //mac 1btn = 0, mac nbutton = 1, linux game descent = 2, windows =3
int lenhelp = 27;
char * keyboardShortcutHelp[27] = {
"EXAMINE Mode",
"   LMB rotation: MX rotation around Y axis; MY rotation around X axis",
"   RMB zooms", // On Apple computers with one button mice, press and hold the "control" key, and use your mouse. 
"WALK Mode",
"   LMB movement: MX left/right turns; MY walk forward/backward", 
"   RMB height", //se Button 3 moves you up/down (changes your height above the ground). On Apple computers with one button mice, press and hold the "control" key, and use your mouse. 
"FLY Mode",
"   8 k rotation down/up",
"   u o rotation left/right",
"   7 9 rotation about the Z axis",
"   a z translation forwards/backwards",
"   j l translation left/right",
"   p ; translation up/down",
"EXFLY Mode",
"   takes input from the file /tmp/inpdev",
"all modes",
"  d Switch to Fly (Keyboard input) navigation mode", 
"  f Switch to Fly (External Sensor input) navigation mode",
"  e Switch to Examine navigation mode",
"  w Switch to Walk navigation mode",
"  v Go to next viewpoint in the scene",
"  b Go to previous viewpoint in the scene",
"  / Print current viewport local pose", 
"  h Toggle headlight",
"  c Toggle collision detection",
"  x Snapshot",
"  q Quit browser"
};
void printKeyboardHelp(ppstatusbar p)
{
	int j; 
	for(j=0;j<lenhelp;j++)
	{
		XY xy = text2screen(0,j);
		if(p->posType==1)
			FW_GL_RASTERPOS2I(xy.x,xy.y)
		else
			FW_GL_WINDOWPOS2I(xy.x,xy.y)
		printString(keyboardShortcutHelp[j]); 
	}
}

void hudSetConsoleMessage(char *buffer)
{
	s_list_t* last;
	/*calling program keeps ownership of buffer and deletes or recycles buffer*/
	char *buffer2, *line, *ln, *buf;
	int linelen;
	ppstatusbar p = (ppstatusbar)gglobal()->statusbar.prv;

	int len = (int) strlen(buffer)+1;
	buffer2 = malloc(len);
	strncpy(buffer2,buffer,len);
	/* rule: if you have a \n at the end of your buffer, 
	then leave here on a new line, else we'll append to your last line*/
	if(!p->conlist)
	{
		line = malloc(2);
		line[0] = '\0';
		last = ml_new(line);
		p->conlist = last;
		p->concount = 1;
	}
	else
		last = ml_last(p->conlist);
	line = last->elem;
	linelen = (int) strlen(line);
	buf = buffer2;
	do
	{
		ln = strchr(buf,'\n');
		if(ln)
		{ 
			*ln = '\0';
		}
		len = (int) strlen(buf);
		linelen += len + 1;
		line = realloc(line,linelen);
		line = strcat(line,buf); 
		last->elem = line; /* new address from realloc */
		if(ln)
		{	
			*ln = '\n'; /* restore, in case we need it \n */
			buf = &ln[1];
			line = malloc(2);
			line[0] = '\0';
			linelen = (int) strlen(line);
			last = ml_new(line);
			ml_append(p->conlist,last);
			p->concount++;
			if( p->concount > 50 ) // > MAXMESSAGES number of scrolling lines
			{
				//s_list_t* temp;
				free((char*)p->conlist->elem);
				p->conlist = ml_delete_self(p->conlist, p->conlist); /*delete from top*/
				p->concount--;
			}
		}
	}while(ln);
	free(buffer2);
}

void printConsoleText()
{
	/* ConsoleMessage() comes out as a multi-line history rendered over the scene */
	int jstart;
	char* buf;
	int j = 0;
	XY xybottom;
	ppstatusbar p = (ppstatusbar)gglobal()->statusbar.prv;

	jstart = j;
	{
		s_list_t *__l;
		s_list_t *next;
		s_list_t *_list = p->conlist;
		/* lets keep the scrolling text from touching the bottom of the screen */
		xybottom = screen2text(0,0); 
		jstart = max(0,p->concount-(xybottom.y - 3)); /* keep it 3 lines off the bottom */
		for(__l=_list;__l!=NULL;) 
		{
			next = ml_next(__l); /* we need to get next from __l before action deletes element */ 
			if(j >= jstart) /* no need to print off-screen text */
			{
				XY xy = text2screen(0,j-jstart);
				if(p->posType==1)
					FW_GL_RASTERPOS2I(xy.x,xy.y)
				else
					FW_GL_WINDOWPOS2I(xy.x,xy.y)
				buf = __l->elem;
				printString(__l->elem); 
			}
			j++;
			__l = next; 
		}
	}
}

void initButtons()
{
	/* first time renderButtons() is called, this is called to load the button icons */
	int i;
	ttglobal tg = gglobal();
	ppstatusbar p = (ppstatusbar)tg->statusbar.prv;
	tg->Mainloop.clipPlane = 16;

	for(i=0;i<p->nbuts;i++)
	{
		/* compute the statusbar layout rectangles assuming 64x64 icons scaled by .5 to 32x32 */
		p->butrect[0][i] = 5+(i*32);   /* lower left  x */
		p->butrect[1][i] = 0;				/* lower left  y */
		p->butrect[2][i] = 5+(i*32)+32;/* upper right x */
		p->butrect[3][i] = 32;				/* upper right y */
	}
	if(p->buttonType ==0)
	{
		/* png icon files (can have transparency) problem: you need to put them in the current working directory*/
		for(i=0;i<p->nbuts;i++)
		  texture_load_from_file(&p->butts[i][0], butFnames[i]);
		//bool texture_load_from_file(textureTableIndexStruct_s* this_tex, char *filename);
		for(i=0;i<p->nbuts;i++)
		{
			/* compute grayed out (non active) versions */
			int j,k,l,w,h,g,size;
			w = p->butts[i][0].x;
			h = p->butts[i][0].y;
			size = w * h * 4;
			p->iconSize = w;
			p->butts[i][1].texdata = malloc(size);
			memcpy(p->butts[i][1].texdata,p->butts[i][0].texdata,size);
			for(j=0;j<p->butts[i][0].x;j++)
				for(k=0;k<p->butts[i][0].y;k++)
				{
					g = 0;
					for(l=0;l<3;l++)
						g = g + p->butts[i][0].texdata[j*w*4 + k*4 + l];
					g = g / 3;
					for(l=0;l<3;l++)
						p->butts[i][0].texdata[j*w*4 + k*4 + l] = g;
				}
		}
	
		if( p->savePng2dotc )
		{
			/* write rgba out as binary bitmap in .c struct format for inclusion above */
			FILE* out = fopen("buttons_c","w+");
			for(i=0;i<p->nbuts;i++)
			{
				int j,k,l,m,w,h,g,size;
				unsigned char row, a, bit;
				w = p->butts[i][0].x;
				h = p->butts[i][0].y;
				size = w * h * 4;
				fprintf(out,"GLubyte %s[] = {\n",butFnames[i]); 
				for(j=0;j<p->butts[i][0].y;j++)
				{
					for(k=0;k<p->butts[i][0].x;k+=8)
					{
						row = 0;
						for(m=0;m<8;m++)
						{
							g = 0;
							for(l=0;l<3;l++)
								g = g + p->butts[i][0].texdata[j*w*4 + (k+m)*4 + l];
							g = g / 3;
							a = p->butts[i][0].texdata[j*w*4 + (k+m)*4 + 3];
							bit = 0;
							if( a > 127 ) //(float)a/256.0f > .5 ) //simple thresholding at .5 on the alpha channel
								bit = 1;
							row |= bit << (7 - m);
						}
						fprintf(out,"0x%x,",row);
					}
					fprintf(out,"\n");
				}
				fprintf(out,"};\n");
			}
			fclose(out);
		}
	}
	else if(p->buttonType == 1)
	{
		p->bmIcon[0] = walk;
		p->bmIcon[1] = fly;
		p->bmIcon[2] = examine;
		p->bmIcon[3] = level;
		p->bmIcon[4] = headlight;
		p->bmIcon[5] = collision;
		p->bmIcon[6] = prev;
		p->bmIcon[7] = next;
		p->bmIcon[8] = help;
		p->bmIcon[9] = messages;
		p->bmIcon[10] = options;
		p->bmIcon[11] = reload;

	}

	p->butsLoaded = 1;
}
/* the following setMenuButton_ were defined for some other front end and called from various locations
   - re-using here*/
void setMenuButton_collision(int val){	
	ppstatusbar p = (ppstatusbar)gglobal()->statusbar.prv;
	p->butStatus[5] = val; 
}
void setMenuButton_texSize(int size){ 
	/* this isn't called in my configuration so I don't know what the range is */
	printf("text size=%d\n",size);
	//int bmfontsize = 2; /* 0,1 or 2 - our current size range*/
}
void setMenuButton_headlight(int val){ 
	ppstatusbar p = (ppstatusbar)gglobal()->statusbar.prv;
	p->butStatus[4] = val;
}
void setMenuButton_navModes(int type)
{
	int i;
	ppstatusbar p = (ppstatusbar)gglobal()->statusbar.prv;

	for(i=0;i<4;i++)
		p->butStatus[i] = 0;
	switch(type)
	{
		case VIEWER_NONE: break;
		case VIEWER_EXAMINE: {p->butStatus[2] = 1;break;}
		case VIEWER_WALK: {p->butStatus[0] = 1;break;}
		//case VIEWER_EXFLY:{p->butStatus[2] = 1;break;}
		case VIEWER_FLY:{p->butStatus[1] = 1;break;}
	}
return;
}
void handleButtonOver()
{
	/* called from mainloop > fwl_handle_aqua to 
	a) detect a button over and 
	b) highlight underneath the button*/
	int i,x,y;
	ppstatusbar p;
	ttglobal tg = gglobal();
	p = (ppstatusbar)tg->statusbar.prv;

	x = tg->Mainloop.currentX[0];
	y = tg->display.screenHeight - tg->Mainloop.currentY[0];
	p->isOver = -1;
	for(i=0;i<p->nbuts;i++)
		if(x > p->butrect[0][i] && x < p->butrect[2][i] 
		&& y > p->butrect[1][i] && y < p->butrect[3][i] ) 
		{
			/* printf("%d",i); */  /* is over */
			p->isOver = i;
			break;
		}
}
void handleButtonPress()
{
	/* called from mainloop > to 
	a) detect a button hit and 
	b) toggle the button icon and
	c) set the related option
	*/
	int i,j,x,y,oldval;
	ppstatusbar p;
	ttglobal tg = gglobal();
	p = (ppstatusbar)tg->statusbar.prv;

	x = tg->Mainloop.currentX[0];
	y = tg->display.screenHeight - tg->Mainloop.currentY[0];
	for(i=0;i<p->nbuts;i++)
		if(x > p->butrect[0][i] && x < p->butrect[2][i] 
		&& y > p->butrect[1][i] && y < p->butrect[3][i] )
		{
			/* printf("[%d=",i); */  /* button pressed */
			if( i < 3 )
			{
				/* radio button - just one on at a time walk,fly,examine */
				for(j=0;j<3;j++) p->butStatus[j] = 0;
				p->butStatus[i] = 1;
				switch(i)
				{
                    case 0: { fwl_set_viewer_type (VIEWER_WALK); break; }
                    case 1: { fwl_set_viewer_type (VIEWER_FLY); break; }
                    //case 2: { fwl_set_viewer_type (VIEWER_EXFLY); break; }
                    case 2: { fwl_set_viewer_type (VIEWER_EXAMINE); break; }
				}

			}
			else if( i > 7 && i < 11 )
			{
				/* radio button except can toggle all off - at most one on at a time  ?!check*/
				oldval = p->butStatus[i];
				for(j=8;j<11;j++) p->butStatus[j] = 0;
				p->butStatus[i] = 1 - oldval;
				p->showOptions = p->butStatus[10];
			}
			else if(i == 4 || i == 5)
			{
				/* toggle from current state - headlight, collision */
				p->butStatus[i] = 1 - p->butStatus[i];
				switch(i)
				{
                    			case 4: { fwl_toggle_headlight(); break;}
					case 5: { toggle_collision(); break; }
				}
			}
			else if(i == 6 || i == 7 )
			{
				/* prev, next */
				switch(i)
				{
                case 6: {fwl_Prev_ViewPoint(); break;}
                case 7: {fwl_Next_ViewPoint(); break;}
				}
			}
			else if(i == 3)
			{
				/* level */
				//printf("leveling function not implemented\n");
				viewer_level_to_bound();
			}
			else if(i == 11)
			{
				fwl_reload();
			}

			/*printf("%d",butStatus[i]);
			printf("]");*/
			break;
		}
}
void renderButtons()
{
	/* called from drawStatusBar() to render the user buttons like walk/fly, headlight, collision etc. */
	int i,loaded;
	float zoomfactor;
	ppstatusbar p;
	ttglobal tg = gglobal();
	p = (ppstatusbar)tg->statusbar.prv;


	// get rid of compiler warning
	loaded = 0;


	if(!p->butsLoaded)
		initButtons();
	FW_GL_SCISSOR(0,0,tg->display.screenWidth,tg->Mainloop.clipPlane*2);
	FW_GL_ENABLE(GL_SCISSOR_TEST);
	FW_GL_CLEAR_COLOR(.922f,.91f,.844f,1.0f); //windowing gray
	//glClearColor(.754f,.82f,.93f,1.0f); //193.0f/256.0f,210.0f/256.0f,238.0f/256.0f,1.0f); //windowing blue
	FW_GL_CLEAR(GL_COLOR_BUFFER_BIT);
	FW_GL_DISABLE(GL_SCISSOR_TEST);
	doglClearColor(); //set back for other cases
	zoomfactor = (float)p->iconSize/32.0f;
	FW_GL_PIXELZOOM(zoomfactor,zoomfactor); //.5f,.5f);
	for(i=0;i<p->nbuts;i++)
	{
		if(p->buttonType==0) loaded = (p->butts[i][0].status == 2);
		if(p->buttonType==1) loaded = p->butsLoaded;
		if( loaded) // butts[i][0].status == 2)
		{
			if(i==p->isOver)
			{
				/*draw a background highlight rectangle*/
				float rgba[4];
				//rgba[0] = .754f; rgba[1] = .82f; rgba[2] = .93f, rgba[3] = 1.0f;
				rgba[0] = .8f; rgba[1] = .87f; rgba[2] = .97f, rgba[3] = 1.0f;
				if(p->posType==1)
					FW_GL_RASTERPOS2I(p->butrect[0][i],p->butrect[1][i])
				else
					FW_GL_WINDOWPOS2I(p->butrect[0][i],p->butrect[1][i])
				FW_GL_PIXELZOOM((float)(p->butrect[2][i]-p->butrect[0][i]),(float)(p->butrect[3][i]-p->butrect[1][i]));
				FW_GL_DRAWPIXELS(1,1,GL_RGBA,GL_FLOAT,rgba);
				//restore
				FW_GL_PIXELZOOM(zoomfactor,zoomfactor); //.5f,.5f);
			}
			if(p->buttonType==1)
			{
				if(p->butStatus[i]) {
					FW_GL_COLOR3F(.6f,.7f,.8f); //.617f,.793f,.920f); /*light blue*/
				} else {
					FW_GL_COLOR3F(.8f,.8f,.8f); /*light gray*/
				}
			}
			if(p->posType==1)
				FW_GL_RASTERPOS2I(p->butrect[0][i],p->butrect[1][i])
			else
				FW_GL_WINDOWPOS2I(p->butrect[0][i],p->butrect[1][i])
			if(p->buttonType==0)
				FW_GL_DRAWPIXELS(p->butts[i][0].x,p->butts[i][0].y,GL_BGRA,GL_UNSIGNED_BYTE,p->butts[i][p->butStatus[i]].texdata);
			else if(p->buttonType==1)
				FW_GL_BITMAP(32,32, 0.0f, 0.0f, 0.0f, 0.0f, p->bmIcon[i]);

		}
	}
	FW_GL_PIXELZOOM(1.0f,1.0f);
	p->hadString = 1;
}
void setArrowCursor(); //from ui/common.h
void updateCursorStyle();
int handleStatusbarHud(int mev, int* clipplane)
{
	ppstatusbar p;
	ttglobal tg = gglobal();
	p = (ppstatusbar)tg->statusbar.prv;

    if ((mev == ButtonPress) || (mev == ButtonRelease)) 
	{
        /* record which button is down */
		/* >>> statusbar hud */
		if( p->showButtons)
		{
			if(mev==ButtonPress)
				handleButtonPress();
			return 1;
		}
		if(p->showOptions)
		{
			if(mev==ButtonPress)
				handleOptionPress();
			return 1;
		}
	}
    if (mev == MotionNotify) 
	{
		int clipline;
		(*clipplane) = 16;

		/* >>> statusbar hud */
		if((*clipplane) || p->showButtons)
		{
			clipline = *clipplane;
			if(p->showButtons) clipline = 2*(*clipplane);

			if( tg->display.screenHeight - tg->Mainloop.currentY[0] < clipline )
			{
				p->showButtons = 1;
				setArrowCursor();
				handleButtonOver();
				return 1; /* don't process for navigation */
			}
			else
			{
				p->showButtons = 0;
			}
		}
		if(p->showOptions)
		{
			/* let HUD options menu swallow button clicks */
			return 1;
		}
		/* <<< statusbar hud */
	}
	return 0;
}
char *getMessageBar(); //in common.c

void drawStatusBar() 
{
	/* drawStatusBar() is called just before swapbuffers in mainloop so anything that you want to render 2D
	   (non-scene things like browser status messages (like FPS), 
	   browser control option buttons (menu bar) and checkboxes, console error messages)
	   you can put in here.
	   Ideally things like buttons and status could be hidden/optional/configurable, since some
	   applications don't want to give option control to the user - for example a museum kiosk application -
	   and some applications have gui widgets for it.
	
	The interface that statusbarHud implements
	let S be statusbar, M be menubar, C be console and H be options+help
	//previously defined interfaces implemented here
S	int sb_hasString;
SM	int screenWidth,clipPlane;
S	drawStatusBar() - call before swapbuffers in mainloop
S	void update_status(char* msg); //when cursor over sensitive
S	void kill_status (void); //not sure - called from Mainloop L1331 and OpenGL_Utils L770
S	void setMenuFps (float fps);
M	void setMenuButton_collision(int val); //called from mainloop do_KeyPress
M	void setMenuButton_headlight(int val); // "
M	void setMenuButton_navModes(int type); // "
M	void setMenuButton_texSize(int size);  // not called or used in my current config

	//new interfaces for statusbarHud you will need to implement stubs for when not using hud:
C	void setHudConsoleMessage(char *buffer); //call from ConsoleMessageHud.c
H	int handleStatusbarHud(int mev, int* clipplane); //called from fwl_handle_aqua or handle_xevent in mainloop

	The interface statusbarHud requires other modules to implement to serve it:
	//already implemented
MH	int currentX[0],currentY[0] - mouse coords
CH	int screenHeight - in pixels
H	Viewer.(various stereo params)
	//new requirements for statusbarHud:
H	void setStereo(int type);							//implement in viewer.c
H	void toggleOrSetStereo(int type);					// "
H	void setAnaglyphSideColor(char color, int iside);	//"
H	void updateEyehalf();								//" 
M	viewer_level_to_bound();							//"
M       void toggle_collision()                             //"
    */
	char *pp; 
	//float c[4];
	//int ic[4];
	ppstatusbar p;
	ttglobal tg = gglobal();
	p = (ppstatusbar)tg->statusbar.prv;

	tg->ConsoleMessage.Console_writeToHud = 1;
	//Console_writeToCRT = 1;
	//Console_writeToFile = 0;
	FW_GL_DEPTHMASK(GL_FALSE);
	//if(true) //for testing ogl 1.1 and rasterpos (vs 1.4 and windowpos)
	if(!tg->display.rdr_caps.have_GL_VERSION_1_4)
	{
		//p.306 redbook - glwindowpos2i is ogl 1.4, older is glrasterpos2i, and for that
		//you must set up orthomatrix
		FW_GL_VIEWPORT(0, 0, tg->display.screenWidth, tg->display.screenHeight);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluOrtho2D(0.0,tg->display.screenWidth,0.0,tg->display.screenHeight);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glDisable(GL_LIGHTING); // http://www.opengl.org/wiki/Coloring_a_bitmap 
		glDisable(GL_DEPTH_TEST);
		p->posType = 1; // use RasterPos2i instead of WindowPos2i
	}

	if(p->showButtons)
	{
		renderButtons();
		FW_GL_DEPTHMASK(GL_TRUE);
		if(p->posType==1) { glEnable(GL_LIGHTING); glEnable(GL_DEPTH_TEST); }
		return;
	}
	if (!p->sb_hasString && !p->showConText &&!p->butStatus[8] &&!p->butStatus[9] && !p->butStatus[10]) {
		if(p->hadString || !p->initDone || true)
		{
			/* clear the status bar because there's nothing to show */
			if(tg->Mainloop.clipPlane == 0) tg->Mainloop.clipPlane = 16;
			FW_GL_SCISSOR(0,0,tg->display.screenWidth,tg->Mainloop.clipPlane);
			FW_GL_ENABLE(GL_SCISSOR_TEST);
			FW_GL_CLEAR_COLOR(.922f,.91f,.844f,1.0f); //windowing gray
			FW_GL_CLEAR(GL_COLOR_BUFFER_BIT);
			FW_GL_DISABLE(GL_SCISSOR_TEST);
			p->hadString = 0;
			p->initDone = TRUE;
		}
		FW_GL_CLEAR_COLOR(0.0f,0.0f,0.0f,1.0f); 
		FW_GL_DEPTHMASK(GL_TRUE);
		if(p->posType==1) { glEnable(GL_LIGHTING); glEnable(GL_DEPTH_TEST); }
		return;
	}
	/* to improve frame rates we don't need to update the status bar every loop,
	because the mainloop scene rendering should be using a scissor test to avoid glClear()ing 
	the statusbar area. 
	*/
	p->loopcount++;
	if(p->loopcount < 15 && !p->hadString)
	{
		FW_GL_DEPTHMASK(GL_TRUE);
		if(p->posType==1) { glEnable(GL_LIGHTING); glEnable(GL_DEPTH_TEST); }
		return;
	}
	p->loopcount = 0;

	/* OK time to update the status bar */
	if(!p->fontInitialized) initFont();
	/* unconditionally clear the statusbar area */
	FW_GL_SCISSOR(0,0,tg->display.screenWidth,tg->Mainloop.clipPlane);
	FW_GL_ENABLE(GL_SCISSOR_TEST);
	FW_GL_CLEAR_COLOR(.922f,.91f,.844f,1.0f); //windowing gray
	FW_GL_CLEAR(GL_COLOR_BUFFER_BIT);
	FW_GL_DISABLE(GL_SCISSOR_TEST);

	// you must call drawStatusBar() from render() just before swapbuffers 
	FW_GL_DEPTHMASK(FALSE);
	FW_GL_DISABLE(GL_DEPTH_TEST);
	FW_GL_COLOR3F(0.2f,0.2f,0.5f);
	//glWindowPos seems to set the bitmap color correctly in windows
	if(p->posType==1)
		FW_GL_RASTERPOS2I(5,0)
	else
		FW_GL_WINDOWPOS2I(5,0)
	if(p->sb_hasString)
	{
		pp = p->buffer;
		/* print status bar text - things like PLANESENSOR */
		printString(pp); 
		p->hadString = 1;
	}
	//else
	{
		char *strfps,*strstatus;
		//printString(p->messagebar);
		strfps = getMessageBar();
		strstatus = &strfps[15];
		if(p->posType==1)
			FW_GL_RASTERPOS2I(150,0)
		else
			FW_GL_WINDOWPOS2I(150,0) //300,0);
		printString(strfps);
		if(p->posType==1)
			FW_GL_RASTERPOS2I(300,0)
		else
			FW_GL_WINDOWPOS2I(300,0)
		printString(strstatus);
	}
	FW_GL_COLOR3F(1.0f,1.0f,1.0f);
	FW_GL_SHADEMODEL(GL_FLAT);
	if(p->butStatus[8]) printKeyboardHelp(p);
	if(p->butStatus[9]) printConsoleText();
	if(p->butStatus[10]) printOptions();

	FW_GL_CLEAR_COLOR(0.0f,0.0f,0.0f,1.0f); 
	FW_GL_DEPTHMASK(TRUE);
	FW_GL_ENABLE(GL_DEPTH_TEST);
	FW_GL_FLUSH();
	FW_GL_SHADEMODEL(GL_SMOOTH);
	if(p->posType==1) { glEnable(GL_LIGHTING); glEnable(GL_DEPTH_TEST); }
}
#endif
